@@ -10,7 +10,7 @@ use std::time::Duration as StdDuration;
 use junction_typeinfo::TypeInfo;
 
 /// A fraction, expressed as a numerator and a denominator.
-#[derive(Serialize, Deserialize, Clone, Debug, PartialEq)]
+#[derive(Serialize, Deserialize, Clone, Debug, PartialEq, Eq)]
 #[cfg_attr(feature = "typeinfo", derive(TypeInfo))]
 pub struct Fraction {
     pub numerator: i32,
@@ -25,6 +25,14 @@ pub struct Fraction {
 #[derive(Clone)]
 pub struct Regex(regex::Regex);
 
+impl PartialEq for Regex {
+    fn eq(&self, other: &Self) -> bool {
+        self.0.as_str() == other.0.as_str()
+    }
+}
+
+impl Eq for Regex {}
+
 impl std::fmt::Debug for Regex {
     fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> Result<(), std::fmt::Error> {
         f.write_str(self.0.as_str())
@@ -100,12 +108,6 @@ impl<'de> Deserialize<'de> for Regex {
     }
 }
 
-impl PartialEq for Regex {
-    fn eq(&self, other: &Self) -> bool {
-        self.0.as_str() == other.0.as_str()
-    }
-}
-
 /// A wrapper around [std::time::Duration] that serializes to and from a f64
 /// number of seconds.
 #[derive(Copy, Clone, PartialEq, Eq)]
