@@ -21,7 +21,10 @@ use xds_api::pb::{
         },
         r#type::matcher::v3::{string_matcher::MatchPattern, StringMatcher},
     },
-    google::{self, protobuf},
+    google::{
+        self,
+        protobuf::{self, UInt32Value},
+    },
 };
 
 use crate::xds::shared::{parse_xds_regex, regex_matcher};
@@ -230,7 +233,7 @@ impl RouteRule {
         };
 
         let timeouts = RouteTimeouts::from_xds(action)?;
-        let retry = action.retry_policy.as_ref().map(RouteRetry::from_xds);
+        let retry = action.retry_policy.as_ref().and_then(RouteRetry::from_xds);
         // cluster_specifier is a oneof field, so let WeightedTarget specify the
         // field names in errors. we still need to get the index to the
         let backends = BackendRef::from_xds(action.cluster_specifier.as_ref())
@@ -373,7 +376,7 @@ impl RouteMatch {
             }
         }
 
-        // query paramters get their own field name, the oneof happens inside
+        // query parameters get their own field name, the oneof happens inside
         // the array of matches.
         let query_params = r
             .query_parameters
@@ -577,26 +580,36 @@ impl PathMatch {
 }
 
 impl RouteRetry {
-    pub fn from_xds(r: &xds_route::RetryPolicy) -> Self {
-        let codes = r.retriable_status_codes.clone();
-        let attempts = Some(1 + r.num_retries.clone().map_or(0, |v| v.into()));
+    pub fn from_xds(r: &xds_route::RetryPolicy) -> Option<Self> {
+        if r.retriable_status_codes.is_empty()
+            && r.num_retries.is_none()
+            && r.retry_back_off.is_none()
+        {
+            //need to do this as some of the timeout logic is carried in the same proto
+            return None;
+        }
+        let codes = r
+            .retriable_status_codes
+            .iter()
+            .map(|code| *code as u16)
+            .collect();
+        let attempts = r.num_retries.clone().map(|v| u32::from(v) + 1);
         let backoff = r
             .retry_back_off
             .as_ref()
             .and_then(|r2| r2.base_interval.clone().map(|x| x.try_into().unwrap()));
-        Self {
+        Some(Self {
             codes,
             attempts,
             backoff,
-        }
+        })
     }
 
     pub fn to_xds(&self) -> xds_route::RetryPolicy {
-        let retriable_status_codes = self.codes.clone();
+        let retriable_status_codes = self.codes.iter().map(|&code| code as u32).collect();
         let num_retries = self
             .attempts
-            .map(|attempts| attempts.saturating_sub(1))
-            .unwrap_or(0);
+            .map(|attempts| UInt32Value::from(attempts.saturating_sub(1)));
 
         let retry_back_off = self.backoff.map(|b| xds_route::retry_policy::RetryBackOff {
             base_interval: Some(b.try_into().unwrap()),
@@ -605,7 +618,7 @@ impl RouteRetry {
 
         xds_route::RetryPolicy {
             retriable_status_codes,
-            num_retries: Some(num_retries.into()),
+            num_retries,
             retry_back_off,
             ..Default::default()
         }
@@ -731,7 +744,7 @@ impl BackendRef {
                     .collect::<Result<Vec<_>, _>>()
                     .with_fields("weighted_clusters", "clusters")
             }
-            Some(_) => Err(Error::new_static("unsupporetd cluster specifier")),
+            Some(_) => Err(Error::new_static("unsupported cluster specifier")),
             None => Ok(Vec::new()),
         }
     }
@@ -1259,6 +1272,7 @@ mod test {
         });
     }
 
+    #[track_caller]
     fn assert_roundtrip<T, Xds>(v: T)
     where
         T: PartialEq + std::fmt::Debug,
