@@ -106,12 +106,9 @@ async fn main() {
             continue;
         }
 
-        let prod_endpoints = prod_endpoints.unwrap();
-        let staging_endpoints = staging_endpoints.unwrap();
-        let prod = prod_endpoints.first().unwrap();
-        let staging = staging_endpoints.first().unwrap();
-
-        println!("prod={:<20} staging={:<20}", prod.address, staging.address);
+        let prod = prod_endpoints.unwrap();
+        let staging = staging_endpoints.unwrap();
+        println!("prod={:<20} staging={:<20}", prod.addr(), staging.addr());
 
         tokio::time::sleep(Duration::from_millis(1500)).await;
     }
