@@ -1,6 +1,9 @@
 use crate::{
-    endpoints::EndpointIter, load_balancer::BackendLb, xds::AdsClient, ConfigCache, Endpoint,
-    Error, StaticConfig,
+    endpoints::{EndpointGroup, EndpointIter},
+    error::Trace,
+    load_balancer::BackendLb,
+    xds::AdsClient,
+    ConfigCache, Endpoint, Error, StaticConfig,
 };
 use futures::FutureExt;
 use junction_api::{
@@ -8,8 +11,8 @@ use junction_api::{
     http::{HeaderMatch, PathMatch, QueryParamMatch, Route, RouteMatch, RouteRule},
 };
 use rand::distributions::WeightedError;
-use std::sync::Arc;
 use std::time::{Duration, Instant};
+use std::{net::SocketAddr, sync::Arc};
 
 /// An outgoing HTTP Request, before any rewrites or modifications have been
 /// made.
@@ -42,27 +45,103 @@ impl<'a> HttpRequest<'a> {
     }
 }
 
-/// The strategy for udpating Routes, Backends, and Endpoints while resolving.
-///
-/// Resolution mode is orthogonal to client configuration mode - it's possible
-/// to have a dynamic client, but to resolve an individual route by only using
-/// data currently in cache.
-///
-/// See [Client::resolve_routes] and [Client::resolve_endpoint].
-#[derive(Debug, Clone, Copy, PartialEq, Eq)]
-#[non_exhaustive]
-pub enum ResolveMode {
-    /// Resolve configuration with only existing routes and backends, do not
-    /// request new ones over the network.
-    ///
-    /// This mode does not disable push-based updates to existing routes or
-    /// backends. For example, the addresses that are part of a backend or its
-    /// load balancing configuration may still change during resolution.
-    Static,
+/// The result of resolving a route (see [Client::resolve_route]).
+#[derive(Debug, Clone)]
+pub struct ResolvedRoute {
+    /// The resolved route.
+    pub route: Arc<Route>,
 
-    /// Update configuration dynamically as part of making this request. New
-    /// routes, backends, and addresses may fetched to make this request.
-    Dynamic,
+    /// The index of the rule that matched the request.
+    //TODO: doesn't need to be optional? remove it for the request trace anyway
+    pub rule: usize,
+
+    /// The backend selected as part of route resolution.
+    pub backend: BackendId,
+
+    /// smuggle the request trace through here
+    trace: Trace,
+}
+
+/// The context required to select an address from a backend. Includes the URL
+/// and headers from an outgoing request.
+#[derive(Debug, Clone)]
+pub struct LbContext<'a> {
+    url: &'a crate::Url,
+
+    headers: &'a http::HeaderMap,
+
+    previous_addrs: &'a [SocketAddr],
+
+    /// smuggle the request trace through here
+    trace: Trace,
+}
+
+impl<'a> LbContext<'a> {
+    // unused, allowed so that we can make select_endpoint public without exposing Trace
+    #[allow(unused)]
+    pub fn from_parts(url: &'a crate::Url, headers: &'a http::HeaderMap) -> Self {
+        let trace = Trace::new();
+        Self {
+            url,
+            headers,
+            previous_addrs: &[],
+            trace,
+        }
+    }
+
+    fn new(trace: Trace, url: &'a crate::Url, headers: &'a http::HeaderMap) -> Self {
+        Self {
+            url,
+            headers,
+            previous_addrs: &[],
+            trace,
+        }
+    }
+}
+
+/// The result of selecting an endpoint (see [Client::select_endpoint]).
+pub struct SelectedEndpoint {
+    /// The selected endpoint address
+    pub addr: SocketAddr,
+
+    // smuggle trace data back out
+    trace: Trace,
+}
+
+/// The result of making an HTTP request.
+#[derive(Debug, Clone)]
+pub enum HttpResult {
+    /// The client received a complete HTTP response with a status code that was
+    /// not a client error (4xx) or a server error (5xx).
+    StatusOk(http::StatusCode),
+
+    /// The client received a complete HTTP response with a status code that was
+    /// a client error (4xx) or a server error (5xx).
+    StatusError(http::StatusCode),
+
+    /// The client didn't receive a complete HTTP response. This covers any IO
+    /// error or protocol error. From the Junction client's point of view, there
+    /// is no point in distinguishing them.
+    StatusFailed,
+}
+
+impl HttpResult {
+    pub fn is_ok(&self) -> bool {
+        matches!(self, Self::StatusOk(_))
+    }
+
+    pub fn from_u16(code: u16) -> Result<Self, http::status::InvalidStatusCode> {
+        let code = http::StatusCode::from_u16(code)?;
+        Ok(Self::from_code(code))
+    }
+
+    pub fn from_code(code: http::StatusCode) -> Self {
+        if code.is_client_error() || code.is_server_error() {
+            Self::StatusError(code)
+        } else {
+            Self::StatusOk(code)
+        }
+    }
 }
 
 /// A service discovery client that looks up URL information based on URLs,
@@ -80,7 +159,6 @@ pub struct Client {
     // TODO: make configurable with a builder or something, not sure if they
     // will survive.
     resolve_timeout: Duration,
-    resolve_mode: ResolveMode,
 
     config: Config,
 }
@@ -98,7 +176,7 @@ struct DynamicConfig {
     /// a the shared handle to the task that's actually running the client in
     /// the background. should not drop until every active client drops.
     ///
-    /// TOODO: should this get bundled into AdsClient? shrug emoji?
+    /// TODO: should this get bundled into AdsClient? shrug emoji?
     #[allow(unused)]
     ads_task: tokio::task::JoinHandle<()>,
 }
@@ -129,10 +207,7 @@ impl ConfigCache for Config {
         }
     }
 
-    async fn get_endpoints(
-        &self,
-        backend: &BackendId,
-    ) -> Option<Arc<crate::load_balancer::EndpointGroup>> {
+    async fn get_endpoints(&self, backend: &BackendId) -> Option<Arc<EndpointGroup>> {
         match &self {
             Config::Static(s) => s.get_endpoints(backend).await,
             Config::DynamicEndpoints(_, d) => d.ads_client.get_endpoints(backend).await,
@@ -143,7 +218,7 @@ impl ConfigCache for Config {
 
 // FIXME: Vec<Endpoints> is probably the wrong thing to return from all our
 // resolve methods. We probably need a struct that has something like a list
-// of primary endpoints to cycle through on retries, and a seprate list of
+// of primary endpoints to cycle through on retries, and a separate list of
 // endpoints to mirror traffic to. Figure that out once we support mirroring.
 
 impl Client {
@@ -168,11 +243,13 @@ impl Client {
         // once it's started, hand off the task to the executor in the
         // background.
         ads_task.connect().await?;
+
         let handle = tokio::spawn(async move {
-            ads_task.connect().await.expect("xds: connection failed");
             match ads_task.run().await {
                 Ok(()) => (),
-                Err(e) => panic!("xds: ads client exited: unxpected error: {e}"),
+                Err(e) => panic!(
+                    "junction-core: ads client exited with an unexpected error: {e}. this is a bug in Junction!"
+                ),
             }
         });
 
@@ -183,7 +260,6 @@ impl Client {
 
         let client = Self {
             resolve_timeout: Duration::from_secs(5),
-            resolve_mode: ResolveMode::Dynamic,
             config: Config::Dynamic(dyn_config),
         };
 
@@ -206,12 +282,6 @@ impl Client {
             Config::Dynamic(d) => Arc::clone(d),
         };
 
-        dyn_config
-            .ads_client
-            .subscribe_to_backends(static_config.backends())
-            // FIXME: don't panic here
-            .expect("ads is overloaded. this is a bug in Junction");
-
         let config = Config::DynamicEndpoints(static_config, dyn_config);
         Client { config, ..self }
     }
@@ -225,14 +295,139 @@ impl Client {
     /// discovery data.
     pub fn with_static_endpoints(self, routes: Vec<Route>, backends: Vec<Backend>) -> Client {
         let static_config = Arc::new(StaticConfig::with_inferred(routes, backends));
-
-        // FIXME: this needs a way to subscribe to DNS endpoints or specify endpoints
-        // somehow?
-
         let config = Config::Static(static_config);
         Client { config, ..self }
     }
 
+    /// Resolve an HTTP method, URL, and headers into an [Endpoint].
+    ///
+    /// This is the main entry point into Junction. When building an
+    /// integration, use this method to fetch an initial endpoint. After making
+    /// an initial request, use [report_status][Self::report_status] to report
+    /// the status of the request and to retry on failure.
+    ///
+    /// The endpoint returned from this method should be a complete description
+    /// of how to make an HTTP request - it contains the IP address to use, the
+    /// full URL and hostname, the complete set of headers, and retry and timeout
+    /// policy the client should use to make a request.
+    pub async fn resolve_http(
+        &mut self,
+        method: &http::Method,
+        url: &crate::Url,
+        headers: &http::HeaderMap,
+    ) -> crate::Result<Endpoint> {
+        let deadline = Instant::now() + self.resolve_timeout;
+
+        let request = HttpRequest::from_parts(method, url, headers)?;
+        let resolved = self.resolve_route(request, Some(deadline)).await?;
+
+        let lb_context = LbContext::new(resolved.trace, url, headers);
+        let selected = self
+            .select_endpoint(&resolved.backend, lb_context, Some(deadline))
+            .await?;
+
+        let address = selected.addr;
+        let trace = selected.trace;
+        let (timeouts, retry) = {
+            let rule = &resolved.route.rules[resolved.rule];
+            (rule.timeouts.clone(), rule.retry.clone())
+        };
+
+        Ok(Endpoint {
+            method: method.clone(),
+            url: url.clone(),
+            headers: headers.clone(),
+            address,
+            timeouts,
+            retry,
+            backend: resolved.backend,
+            trace,
+            previous_addrs: vec![],
+        })
+    }
+
+    /// Report the status of an externally made HTTP request made against an
+    /// [Endpoint] returned from `resolve_http`.
+    ///
+    /// If retrying the response is appropriate, a new Endpoint will be returned
+    /// with updated address and host info set - calling `resolve_http` to start
+    /// a retry attempt will drop request history and may result in too many
+    /// retries.
+    ///
+    /// If a retry is not appropriate, the returned Endpoint will have updated
+    /// history information, but request details will remain the same. Clients
+    /// may use that value for status or error reporting.
+    pub async fn report_status(
+        &mut self,
+        endpoint: Endpoint,
+        response: HttpResult,
+    ) -> crate::Result<Endpoint> {
+        // TODO: track response stats
+
+        // if there's no reason to pick a new endpoint, just return the existing one as-is
+        if response.is_ok() || !endpoint.should_retry(response) {
+            return Ok(endpoint);
+        }
+
+        // redo endpoint selection
+        // FIXME: real deadline here
+        let deadline = Instant::now() + self.resolve_timeout;
+        let lb_context = LbContext {
+            url: &endpoint.url,
+            headers: &endpoint.headers,
+            previous_addrs: &endpoint.previous_addrs,
+            trace: endpoint.trace,
+        };
+        let next = self
+            .select_endpoint(&endpoint.backend, lb_context, Some(deadline))
+            .await?;
+        let address = next.addr;
+        let trace = next.trace;
+
+        // track address history
+        let mut previous_addrs = endpoint.previous_addrs;
+        previous_addrs.push(endpoint.address);
+
+        Ok(Endpoint {
+            address,
+            trace,
+            previous_addrs,
+            ..endpoint
+        })
+    }
+
+    /// Resolve an HTTP method, URL, and headers to a target backend, returning
+    /// the Route that matched, the index of the rule that matched, and the
+    /// backend that was chosen - to make backend choice determinstic with
+    /// multiple backends, set the `JUNCTION_SEED` environment variable.
+    ///
+    /// This is a lower-level method that only performs the Route matching part
+    /// of resolution. It's intended for debugging or querying a client for
+    /// specific information. For everyday use, prefer [Client::resolve_http].
+    pub async fn resolve_route(
+        &self,
+        request: HttpRequest<'_>,
+        deadline: Option<Instant>,
+    ) -> crate::Result<ResolvedRoute> {
+        let trace = Trace::new();
+        resolve_routes(&self.config, trace, request, deadline).await
+    }
+
+    /// Select an endpoint address for this backend from the set of currently
+    /// available endpoints.
+    ///
+    /// This is a lower level method that only performs part of route
+    /// resolution, and is intended for debugging and testing. For everyday use,
+    /// prefer [Client::resolve_http].
+    pub async fn select_endpoint(
+        &self,
+        backend: &BackendId,
+        ctx: LbContext<'_>,
+        deadline: Option<Instant>,
+    ) -> crate::Result<SelectedEndpoint> {
+        select_endpoint(&self.config, backend, ctx, deadline).await
+    }
+
     /// Start a gRPC CSDS server on the given port. To run the server, you must
     /// `await` this future.
     ///
@@ -256,6 +451,19 @@ impl Client {
         }
     }
 
+    /// Dump xDS resources that failed to update. This is a view of the data
+    /// returned by [Client::dump_xds] that only contains resources with
+    /// errors.
+    pub fn dump_xds_errors(&self) -> Vec<crate::XdsConfig> {
+        match self.config.ads() {
+            Some(ads) => ads
+                .iter_xds()
+                .filter(|xds| xds.last_error.is_some())
+                .collect(),
+            None => Vec::new(),
+        }
+    }
+
     /// Dump the Client's current table of [Route]s, merging together any
     /// default routes and remotely fetched routes the same way the client would
     /// when resolving endpoints.
@@ -278,60 +486,6 @@ impl Client {
         }
     }
 
-    /// Resolve an HTTP method, URL, and headers to a target backend, returning
-    /// the Route that matched, the index of the rule that matched, and the
-    /// backend that was chosen - to make backend choice determinstic with
-    /// multiple backends, set the `JUNCTION_SEED` environment variable.
-    ///
-    /// This is a lower-level method that only performs the Route matching half
-    /// of resolution. It's intended for debugging or querying a client for
-    /// specific information. For everyday use, prefer [Client::resolve_http].
-    /// [Client::resolve_http].
-    pub async fn resolve_routes(
-        &self,
-        config_mode: ResolveMode,
-        request: HttpRequest<'_>,
-    ) -> crate::Result<ResolvedRoute> {
-        if let (ResolveMode::Dynamic, Config::Dynamic(d)) = (config_mode, &self.config) {
-            d.ads_client
-                .subscribe_to_hosts([request.url.authority().to_string()])
-                // FIXME: shouldn't be a panic, should be async
-                .expect("ads client overloaded, this is a bug in Junction");
-        }
-
-        resolve_routes(&self.config, request).await
-    }
-
-    /// Use a resolved Route and Backend to select an endpoint for a request.
-    ///
-    /// This is a lower level method that only performs the Backend
-    /// load-balancing half of resolution. It's intended for debugging or
-    /// querying a client for specific information. For everday
-    /// use, prefer [Client::resolve_http].
-    pub async fn resolve_endpoint(
-        &self,
-        resolve_mode: ResolveMode,
-        resolved: ResolvedRoute,
-        request: HttpRequest<'_>,
-    ) -> crate::Result<Vec<Endpoint>> {
-        // subscribe to backends with either dynamic endpoints or full
-        // dynamic mode, since this is how we'll fetch endpoints.
-        //
-        // TODO: should there be a way to subscribe to endpoints here as well?
-        // TODO: should we just get rid of the subscribe here and move it into
-        //       the get_routes/get_backends/get_endpoints calls???
-        if let (ResolveMode::Dynamic, Config::Dynamic(d) | Config::DynamicEndpoints(_, d)) =
-            (resolve_mode, &self.config)
-        {
-            d.ads_client
-                .subscribe_to_backends([resolved.backend.clone()])
-                // FIXME: shouldn't be a panic, should be async
-                .expect("ads client overloaded, this is a bug in Junction");
-        }
-
-        resolve_endpoint(&self.config, resolved, request).await
-    }
-
     /// Return the endpoints currently in cache for this backend.
     ///
     /// The returned endpoints are a snapshot of what is currently in cache and
@@ -343,163 +497,166 @@ impl Client {
             .flatten()
             .map(EndpointIter::from)
     }
-
-    /// Resolve an HTTP method, URL, and headers into a set of [Endpoint]s.
-    ///
-    /// When multiple endpoints are returned, a client should send traffic to
-    /// ALL of the returned endpoints because the routing policy specified
-    /// that traffic should be mirrored.
-    pub async fn resolve_http(
-        &mut self,
-        method: &http::Method,
-        url: &crate::Url,
-        headers: &http::HeaderMap,
-    ) -> crate::Result<Vec<crate::Endpoint>> {
-        let request = HttpRequest::from_parts(method, url, headers)?;
-        let deadline = Instant::now() + self.resolve_timeout;
-
-        let resolve_routes = self.resolve_routes(self.resolve_mode, request.clone());
-        let resolved = match tokio::time::timeout_at(deadline.into(), resolve_routes).await {
-            Ok(Ok(r)) => r,
-            Ok(Err(e)) => return Err(e),
-            Err(_) => return Err(Error::timed_out("timed out fetching routes")),
-        };
-
-        let resolve_endpoint = self.resolve_endpoint(self.resolve_mode, resolved, request);
-        match tokio::time::timeout_at(deadline.into(), resolve_endpoint).await {
-            Ok(Ok(r)) => Ok(r),
-            Ok(Err(e)) => Err(e),
-            Err(_) => Err(Error::timed_out("timed out fetching backends")),
-        }
-    }
 }
 
-/// The result of [resolving a route][Client::resolve_routes].
-#[derive(Debug, Clone)]
-pub struct ResolvedRoute {
-    /// The resolved route.
-    pub route: Arc<Route>,
+macro_rules! with_deadline {
+    ($fut:expr, $deadline:expr, $msg:expr, $trace:expr $(,)*) => {
+        tokio::select! {
+            biased;
 
-    /// The index of the rule that matched the request.
-    //TODO: doesn't need to be optional? remove it for the request trace anyway
-    pub rule: Option<usize>,
-
-    /// The backend selected as part of route resolution.
-    pub backend: BackendId,
+            res = $fut => res,
+            _ = sleep_until($deadline) => {
+                return Err(Error::timed_out($msg, $trace));
+            }
+        }
+    };
 }
 
 pub(crate) async fn resolve_routes(
     cache: &impl ConfigCache,
+    mut trace: Trace,
     request: HttpRequest<'_>,
+    deadline: Option<Instant>,
 ) -> crate::Result<ResolvedRoute> {
     use rand::seq::SliceRandom;
 
     // look up the route, grabbing the first route that returns from cache.
-    let matching_route = match cache.get_route(request.url.authority()).await {
-        Some(route) => route,
-        None => return Err(Error::no_route_matched(request.url.authority().to_string())),
+    let route = with_deadline!(
+        cache.get_route(request.url.authority()),
+        deadline,
+        "timed out fetching route",
+        trace,
+    );
+    let Some(route) = route else {
+        return Err(Error::no_route_matched(
+            request.url.authority().to_string(),
+            trace,
+        ));
     };
+    trace.lookup_route(&route);
 
-    // if we got here, we have resolved to a list of routes
-    let Some((matching_rule_idx, matching_rule)) = find_matching_rule(
-        &matching_route,
-        request.method,
-        request.url,
-        request.headers,
-    ) else {
-        return Err(Error::no_rule_matched(matching_route.id.clone()));
+    // match the request against the list of RouteRules that are part of this
+    // request. the hostname and port of the request have already matched but we
+    // need to match headers/url params/method and so on.
+    let (rule, matching_rule) = match find_matching_rule(&route, request.clone()) {
+        Some((idx, r)) => (idx, r),
+        None => return Err(Error::no_rule_matched(route.id.clone(), trace)),
     };
+    trace.matched_rule(
+        rule,
+        route.rules.get(rule).and_then(|rule| rule.name.as_ref()),
+    );
 
-    // pick a target at random from the list, respecting weights.
+    // pick a target at random from the list, respecting weights. if there are
+    // no backends listed we should blackhole here.
     let weighted_backend = &crate::rand::with_thread_rng(|rng| {
         matching_rule.backends.choose_weighted(rng, |wc| wc.weight)
     });
-
     let backend_ref = match weighted_backend {
         Ok(backend_ref) => backend_ref,
         Err(WeightedError::NoItem) => {
+            // TODO: should this just return a special endpoint that 500s?
             return Err(Error::invalid_route(
                 "route has no backends",
-                matching_route.id.clone(),
-                matching_rule_idx,
-            ))
+                route.id.clone(),
+                rule,
+                trace,
+            ));
         }
         Err(_) => {
             return Err(Error::invalid_route(
                 "backends weights are invalid: total weights must be greater than zero",
-                matching_route.id.clone(),
-                matching_rule_idx,
+                route.id.clone(),
+                rule,
+                trace,
             ))
         }
     };
-
     let backend = backend_ref.into_backend_id(request.url.default_port());
+    trace.select_backend(&backend);
+
     Ok(ResolvedRoute {
-        route: matching_route.clone(),
-        rule: Some(matching_rule_idx),
+        route: route.clone(),
+        rule,
         backend,
+        trace,
     })
 }
 
-async fn resolve_endpoint(
+async fn select_endpoint(
     cache: &impl ConfigCache,
-    resolved: ResolvedRoute,
-    request: HttpRequest<'_>,
-) -> crate::Result<Vec<Endpoint>> {
-    let Some(backend) = cache.get_backend(&resolved.backend).await else {
-        return Err(Error::no_backend(
-            resolved.route.id.clone(),
-            resolved.rule,
-            resolved.backend,
-        ));
+    backend: &BackendId,
+    mut ctx: LbContext<'_>,
+    deadline: Option<Instant>,
+) -> crate::Result<SelectedEndpoint> {
+    // start the next trace phase
+    ctx.trace.start_endpoint_selection();
+
+    // lookup backend and endpoints
+    //
+    // these lookups are done sequentially, even though they could be raced, so
+    // that we can tell what step a lookup timed out on. we're assuming that
+    // this is okay because in the background fetching a backend for a cache
+    // will also trigger fetching its endpoints and parallelism here won't do
+    // much for us.
+    let blb = with_deadline!(
+        cache.get_backend(backend),
+        deadline,
+        "timed out fetching backend",
+        ctx.trace,
+    );
+    let Some(blb) = blb else {
+        return Err(Error::no_backend(backend.clone(), ctx.trace));
+    };
+    ctx.trace.lookup_backend(backend);
+
+    let endpoints = with_deadline!(
+        cache.get_endpoints(backend),
+        deadline,
+        "timed out fetching endpoints",
+        ctx.trace,
+    );
+    let Some(endpoints) = endpoints else {
+        return Err(Error::no_reachable_endpoints(backend.clone(), ctx.trace));
+    };
+    ctx.trace.lookup_endpoints(backend);
+
+    // load balance.
+    //
+    // no trace is done here, the load balancer impls stamp the traces themselves
+    let addr = blb.load_balancer.load_balance(
+        &mut ctx.trace,
+        &endpoints,
+        ctx.url,
+        ctx.headers,
+        ctx.previous_addrs,
+    );
+    let Some(addr) = addr else {
+        return Err(Error::no_reachable_endpoints(backend.clone(), ctx.trace));
     };
 
-    let Some(endpoints) = cache.get_endpoints(&resolved.backend).await else {
-        return Err(Error::no_reachable_endpoints(
-            resolved.route.id.clone(),
-            backend.config.id.clone(),
-        ));
-    };
+    Ok(SelectedEndpoint {
+        addr: *addr,
+        trace: ctx.trace,
+    })
+}
 
-    let endpoint = backend
-        .load_balancer
-        .load_balance(request.url, request.headers, &endpoints);
-    let Some(endpoint) = endpoint else {
-        return Err(Error::no_reachable_endpoints(
-            resolved.route.id.clone(),
-            resolved.backend,
-        ));
-    };
-
-    let url = request.url.clone();
-    let (timeouts, retry) = match resolved.rule {
-        Some(idx) => {
-            let rule = &resolved.route.rules[idx];
-            (rule.timeouts.clone(), rule.retry.clone())
-        }
-        None => (None, None),
-    };
-
-    Ok(vec![crate::Endpoint {
-        url,
-        timeouts,
-        retry,
-        address: endpoint.clone(),
-    }])
+async fn sleep_until(deadline: Option<Instant>) {
+    if let Some(d) = deadline {
+        tokio::time::sleep_until(d.into()).await;
+    }
 }
 
 //FIXME(routing): picking between these is way more complicated than finding the
 //first match
 fn find_matching_rule<'a>(
     route: &'a Route,
-    method: &http::Method,
-    url: &crate::Url,
-    headers: &http::HeaderMap,
+    request: HttpRequest<'_>,
 ) -> Option<(usize, &'a RouteRule)> {
     let rule_idx = route
         .rules
         .iter()
-        .position(|rule| is_route_rule_match(rule, method, url, headers))?;
+        .position(|rule| is_route_rule_match(rule, request.method, request.url, request.headers))?;
 
     let rule = &route.rules[rule_idx];
     Some((rule_idx, rule))
@@ -591,7 +748,7 @@ mod test {
 
     #[track_caller]
     fn assert_resolve_routes(cache: &impl ConfigCache, request: HttpRequest<'_>) -> ResolvedRoute {
-        resolve_routes(cache, request)
+        resolve_routes(cache, Trace::new(), request, None)
             .now_or_never()
             .unwrap()
             .unwrap()
@@ -599,7 +756,7 @@ mod test {
 
     #[track_caller]
     fn assert_resolve_err(cache: &impl ConfigCache, request: HttpRequest<'_>) -> crate::Error {
-        resolve_routes(cache, request)
+        resolve_routes(cache, Trace::new(), request, None)
             .now_or_never()
             .unwrap()
             .unwrap_err()
@@ -653,9 +810,7 @@ mod test {
         let request = HttpRequest::from_parts(&http::Method::GET, &url, &headers).unwrap();
 
         let err = assert_resolve_err(&routes, request);
-        assert!(err
-            .to_string()
-            .contains("no routing rules matched the request"));
+        assert!(err.to_string().contains("no rules matched the request"));
         assert!(!err.is_temporary());
     }
 
@@ -735,7 +890,7 @@ mod test {
         let resolved = assert_resolve_routes(&routes, request);
 
         // should match the fallthrough rule
-        assert_eq!(resolved.rule, Some(1));
+        assert_eq!(resolved.rule, 1);
         assert_eq!(resolved.backend, backend_two.as_backend_id(8919));
 
         let url = Url::from_str("http://example.com/users/123").unwrap();
@@ -745,9 +900,7 @@ mod test {
 
         // should match the first rule, with the path match
         assert_eq!(resolved.backend, backend_one.as_backend_id(8910));
-        assert!(!resolved.route.rules[resolved.rule.unwrap()]
-            .matches
-            .is_empty());
+        assert!(!resolved.route.rules[resolved.rule].matches.is_empty());
 
         let url = Url::from_str("http://example.com/users/123").unwrap();
         let headers = &http::HeaderMap::default();
@@ -755,7 +908,7 @@ mod test {
 
         let resolved = assert_resolve_routes(&routes, request);
         // should match the first rule, with the path match
-        assert_eq!(resolved.rule, Some(0));
+        assert_eq!(resolved.rule, 0);
         assert_eq!(resolved.backend, backend_one.as_backend_id(8910));
     }
 
@@ -820,7 +973,7 @@ mod test {
 
             let resolved = assert_resolve_routes(&routes, request);
             // should match the fallthrough rule
-            assert_eq!(resolved.rule, Some(1));
+            assert_eq!(resolved.rule, 1);
             assert_eq!(resolved.backend, backend_two.as_backend_id(8919));
         }
 
@@ -839,7 +992,7 @@ mod test {
             // should match one of the query matches
             assert_eq!(
                 (resolved.rule, &resolved.backend),
-                (Some(0), &backend_one.as_backend_id(8910)),
+                (0, &backend_one.as_backend_id(8910)),
                 "should match the first rule: {url}"
             );
         }
