@@ -2,7 +2,7 @@ use junction_api::{
     backend::{Backend, BackendId},
     http::Route,
 };
-use junction_core::{ResolveMode, ResourceVersion};
+use junction_core::{HttpResult, ResourceVersion};
 use once_cell::sync::Lazy;
 use pyo3::{
     exceptions::{PyRuntimeError, PyValueError},
@@ -13,12 +13,20 @@ use pyo3::{
     wrap_pyfunction, Bound, Py, PyAny, PyResult, Python,
 };
 use serde::Serialize;
-use std::{net::IpAddr, str::FromStr};
+use std::{
+    net::IpAddr,
+    str::FromStr,
+    time::{Duration, Instant},
+};
 use xds_api::pb::google::protobuf;
 
+mod runtime;
+
 #[pymodule]
 fn junction(m: &Bound<'_, PyModule>) -> PyResult<()> {
     m.add_class::<Junction>()?;
+    m.add_class::<Endpoint>()?;
+    m.add_class::<RetryPolicy>()?;
     m.add_function(wrap_pyfunction!(default_client, m)?)?;
     m.add_function(wrap_pyfunction!(check_route, m)?)?;
     m.add_function(wrap_pyfunction!(dump_kube_route, m)?)?;
@@ -49,26 +57,13 @@ mod env {
 /// An endpoint that an HTTP call can be made to. Includes the address that the
 /// request should resolve to along with the original request URI, the scheme to
 /// use, and the hostname to use for TLS if appropriate.
+//
+// TODO: add http method and shadowing. we can't switch method right now so
+// method is moot.
 #[derive(Clone, Debug)]
 #[pyclass]
 pub struct Endpoint {
-    #[pyo3(get)]
-    addr: EndpointAddress,
-
-    #[pyo3(get)]
-    scheme: String,
-
-    #[pyo3(get)]
-    host: String,
-
-    #[pyo3(get)]
-    request_uri: String,
-
-    #[pyo3(get)]
-    retry_policy: Option<RetryPolicy>,
-
-    #[pyo3(get)]
-    timeout_policy: Option<TimeoutPolicy>,
+    inner: junction_core::Endpoint,
 }
 
 #[pymethods]
@@ -76,85 +71,49 @@ impl Endpoint {
     fn __repr__(&self) -> String {
         format!(
             "Endpoint({addr}, {uri})",
-            addr = self.addr,
-            uri = self.request_uri
+            addr = self.inner.addr(),
+            uri = self.inner.url(),
         )
     }
-}
 
-/// An endpoint address. An address can either be an IPAddress or a DNS name,
-/// but will always include a port.
-#[derive(Debug, Clone)]
-#[pyclass]
-enum EndpointAddress {
-    SocketAddr { addr: IpAddr, port: u32 },
-    DnsName { name: String, port: u32 },
-}
-
-#[pymethods]
-impl EndpointAddress {
-    fn __repr__(&self) -> String {
-        self.to_string()
+    #[getter]
+    fn scheme(&self) -> &str {
+        self.inner.url().scheme()
     }
 
-    fn __str__(&self) -> String {
-        self.to_string()
+    #[getter]
+    fn addr(&self) -> IpAddr {
+        self.inner.addr().ip()
     }
-}
 
-impl std::fmt::Display for EndpointAddress {
-    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
-        match self {
-            EndpointAddress::SocketAddr { addr, port } => write!(f, "{addr}:{port}"),
-            EndpointAddress::DnsName { name, port } => write!(f, "{name}:{port}"),
-        }
+    #[getter]
+    fn port(&self) -> u16 {
+        self.inner.addr().port()
     }
-}
 
-impl From<&junction_core::EndpointAddress> for EndpointAddress {
-    fn from(addr: &junction_core::EndpointAddress) -> Self {
-        match addr {
-            junction_core::EndpointAddress::SocketAddr(addr) => Self::SocketAddr {
-                addr: addr.ip(),
-                port: addr.port() as u32,
-            },
-            junction_core::EndpointAddress::DnsName(name, port) => Self::DnsName {
-                name: name.clone(),
-                port: *port,
-            },
-        }
+    #[getter]
+    fn hostname(&self) -> &str {
+        self.inner.url().hostname()
     }
-}
 
-impl From<junction_core::EndpointAddress> for EndpointAddress {
-    fn from(addr: junction_core::EndpointAddress) -> Self {
-        match addr {
-            junction_core::EndpointAddress::SocketAddr(addr) => Self::SocketAddr {
-                addr: addr.ip(),
-                port: addr.port() as u32,
-            },
-            junction_core::EndpointAddress::DnsName(name, port) => Self::DnsName { name, port },
-        }
+    #[getter]
+    fn retry_policy(&self) -> Option<RetryPolicy> {
+        self.inner.retry().clone().map(|r| r.into())
+    }
+
+    #[getter]
+    fn timeout_policy(&self) -> Option<TimeoutPolicy> {
+        self.inner.timeouts().clone().map(|t| t.into())
+    }
+
+    fn print_trace(&self) {
+        self.inner.print_trace();
     }
 }
 
 impl From<junction_core::Endpoint> for Endpoint {
-    fn from(ep: junction_core::Endpoint) -> Self {
-        let scheme = ep.url.scheme().to_string();
-        let host = ep.url.hostname().to_string();
-        let request_uri = ep.url.to_string();
-        let addr = ep.address.into();
-        let retry_policy = ep.retry.map(|r| r.into());
-        let timeout_policy = ep.timeouts.map(|r| r.into());
-
-        Self {
-            scheme,
-            host,
-            request_uri,
-            addr,
-            retry_policy,
-            timeout_policy,
-        }
+    fn from(inner: junction_core::Endpoint) -> Self {
+        Self { inner }
     }
 }
 
@@ -167,7 +126,7 @@ impl From<junction_core::Endpoint> for Endpoint {
 #[pyclass]
 pub struct RetryPolicy {
     #[pyo3(get)]
-    codes: Vec<u32>,
+    codes: Vec<u16>,
 
     #[pyo3(get)]
     attempts: u32,
@@ -176,19 +135,17 @@ pub struct RetryPolicy {
     backoff: f64,
 }
 
-/// A policy that describes how a client should do timeouts.
-#[derive(Clone, Debug)]
-#[pyclass]
-pub struct TimeoutPolicy {
-    #[pyo3(get)]
-    backend_request: f64,
-
-    #[pyo3(get)]
-    request: f64,
-}
-
 #[pymethods]
 impl RetryPolicy {
+    #[new]
+    fn new(codes: Option<Vec<u16>>, attempts: Option<u32>, backoff: Option<f64>) -> Self {
+        Self {
+            codes: codes.unwrap_or_default(),
+            attempts: attempts.unwrap_or_default(),
+            backoff: backoff.unwrap_or_default(),
+        }
+    }
+
     fn __repr__(&self) -> String {
         format!(
             "RetryPolicy({attempts}, {min})",
@@ -209,6 +166,17 @@ impl From<junction_api::http::RouteRetry> for RetryPolicy {
     }
 }
 
+/// A policy that describes how a client should do timeouts.
+#[derive(Clone, Debug)]
+#[pyclass]
+pub struct TimeoutPolicy {
+    #[pyo3(get)]
+    backend_request: f64,
+
+    #[pyo3(get)]
+    request: f64,
+}
+
 #[pymethods]
 impl TimeoutPolicy {
     fn __repr__(&self) -> String {
@@ -232,17 +200,6 @@ impl From<junction_api::http::RouteTimeouts> for TimeoutPolicy {
     }
 }
 
-static RUNTIME: once_cell::sync::Lazy<tokio::runtime::Runtime> = Lazy::new(|| {
-    let rt = tokio::runtime::Builder::new_multi_thread()
-        .worker_threads(2)
-        .enable_all()
-        .thread_name("junction")
-        .build()
-        .expect("failed to initialize a junction runtime");
-
-    rt
-});
-
 /// Check route resolution.
 ///
 /// Resolve a request against a routing table. Returns the full route that was
@@ -259,7 +216,7 @@ fn check_route(
     method: &str,
     url: &str,
     headers: &Bound<PyMapping>,
-) -> PyResult<(Py<PyAny>, Option<usize>, Py<PyAny>)> {
+) -> PyResult<(Py<PyAny>, usize, Py<PyAny>)> {
     let url: junction_core::Url = url
         .parse()
         .map_err(|e| PyValueError::new_err(format!("{e}")))?;
@@ -332,18 +289,13 @@ fn new_client(
     node_name: String,
     cluster_name: String,
 ) -> PyResult<junction_core::Client> {
-    let rt = &RUNTIME;
-    rt.block_on(junction_core::Client::build(
-        ads_address,
-        node_name,
-        cluster_name,
-    ))
-    .map_err(|e| {
-        let error_message = match e.source() {
-            Some(cause) => format!("ads connection failed: {e}: {cause}"),
-            None => format!("ads connection failed: {e}"),
-        };
-        PyRuntimeError::new_err(error_message)
+    runtime::block_and_check_signals(async {
+        junction_core::Client::build(ads_address, node_name, cluster_name)
+            .await
+            .map_err(|e| match e.source() {
+                Some(cause) => format!("ads connection failed: {e}: {cause}"),
+                None => format!("ads connection failed: {e}"),
+            })
     })
 }
 
@@ -431,31 +383,26 @@ impl Junction {
     ///
     /// If `dynamic=False` is passed as a kwarg, the resolution happens without
     /// fetching any new routing data over the network.
-    #[pyo3(signature = (method, url, headers, dynamic=true))]
+    #[pyo3(signature = (method, url, headers, timeout=None))]
     fn resolve_route(
         &mut self,
         py: Python<'_>,
         method: &str,
         url: &str,
         headers: &Bound<PyMapping>,
-        dynamic: bool,
-    ) -> PyResult<(Py<PyAny>, Option<usize>, Py<PyAny>)> {
+        timeout: Option<u64>,
+    ) -> PyResult<(Py<PyAny>, usize, Py<PyAny>)> {
         let method = method_from_py(method)?;
         let url =
             junction_core::Url::from_str(url).map_err(|e| PyValueError::new_err(format!("{e}")))?;
         let headers = headers_from_py(headers)?;
 
+        let deadline = timeout.map(|d| Instant::now() + Duration::from_secs(d));
+
         let request = junction_core::HttpRequest::from_parts(&method, &url, &headers)
             .map_err(|e| PyValueError::new_err(e.to_string()))?;
-
-        let config_mode = match dynamic {
-            true => ResolveMode::Dynamic,
-            false => ResolveMode::Static,
-        };
-
-        let resolved = RUNTIME
-            .block_on(self.core.resolve_routes(config_mode, request))
-            .map_err(|e| PyRuntimeError::new_err(format!("failed to resolve: {e}")))?;
+        let resolved =
+            runtime::block_and_check_signals(self.core.resolve_route(request, deadline))?;
 
         let route = pythonize::pythonize(py, &resolved.route)?;
         let backend = pythonize::pythonize(py, &resolved.backend)?;
@@ -465,14 +412,14 @@ impl Junction {
     /// Return the list of addresses currently in cache for a backend. These
     /// endpoints are a snapshot of what is currently in cache.
     #[pyo3(signature = (backend))]
-    fn get_endpoints(&self, backend: Bound<'_, PyAny>) -> PyResult<Vec<EndpointAddress>> {
+    fn get_endpoints(&self, backend: Bound<'_, PyAny>) -> PyResult<Vec<(IpAddr, u16)>> {
         let backend: BackendId = pythonize::depythonize_bound(backend)?;
         let endpoint_iter = match self.core.get_endpoints(&backend) {
             Some(iter) => iter,
             None => return Ok(Vec::new()),
         };
 
-        Ok(endpoint_iter.addrs().map(EndpointAddress::from).collect())
+        Ok(endpoint_iter.addrs().map(|a| (a.ip(), a.port())).collect())
     }
 
     /// Resolve an endpoint based on an HTTP method, url, and headers.
@@ -487,18 +434,40 @@ impl Junction {
         method: &str,
         url: &str,
         headers: &Bound<PyMapping>,
-    ) -> PyResult<Vec<Endpoint>> {
+    ) -> PyResult<Endpoint> {
         let url =
             junction_core::Url::from_str(url).map_err(|e| PyValueError::new_err(format!("{e}")))?;
         let method = method_from_py(method)?;
         let headers = headers_from_py(headers)?;
 
-        let endpoints = RUNTIME
-            .block_on(self.core.resolve_http(&method, &url, &headers))
-            .map(|endpoints| endpoints.into_iter().map(|e| e.into()).collect())
-            .map_err(|e| PyRuntimeError::new_err(format!("failed to resolve: {e}")))?;
+        let endpoint =
+            runtime::block_and_check_signals(self.core.resolve_http(&method, &url, &headers))?;
 
-        Ok(endpoints)
+        Ok(endpoint.into())
+    }
+
+    #[pyo3(signature = (*, endpoint, status_code=None, error=None))]
+    fn report_status(
+        &mut self,
+        endpoint: Endpoint,
+        status_code: Option<u16>,
+        error: Option<Bound<PyAny>>,
+    ) -> PyResult<Endpoint> {
+        let result = match (status_code, error) {
+            (Some(code), _) => HttpResult::from_u16(code)
+                .map_err(|_| PyValueError::new_err("invalid status code"))?,
+            (None, Some(_)) => HttpResult::StatusFailed,
+            (None, None) => {
+                return Err(PyValueError::new_err(
+                    "either status_code or error is required",
+                ))
+            }
+        };
+
+        let endpoint =
+            runtime::block_and_check_signals(self.core.report_status(endpoint.inner, result))?;
+
+        Ok(endpoint.into())
     }
 
     /// Spawn a new CSDS server on the given port. Spawning the server will not
@@ -507,7 +476,7 @@ impl Junction {
         let run_server = self.core.clone().csds_server(port);
         // FIXME: figure out how to report an error better than this. just
         // printing the exception is good buuuuuuut.
-        RUNTIME.spawn(async move {
+        runtime::spawn(async move {
             if let Err(e) = run_server.await {
                 let py_err = PyRuntimeError::new_err(format!("csds server exited: {e}"));
                 Python::with_gil(|py| py_err.print(py));
@@ -559,6 +528,22 @@ impl Junction {
 
         Ok(values)
     }
+
+    /// Dump the client's current xDS errors as a pbjson dict.
+    ///
+    /// This is the same as dumping config with dump_xds and filtering to only
+    /// xds with a `last_error` message set.
+    fn dump_xds_erorrs(&self, py: Python<'_>) -> PyResult<Vec<Py<PyAny>>> {
+        let mut values = vec![];
+
+        for config in self.core.dump_xds_errors() {
+            let config: XdsConfig = config.into();
+            let as_py = pythonize::pythonize(py, &config)?;
+            values.push(as_py);
+        }
+
+        Ok(values)
+    }
 }
 
 #[derive(Debug, Serialize)]
